/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    CostSensitiveClassifierSplitEvaluator.java
 *    Copyright (C) 2002 University of Waikato
 *
 */


package weka.experiment;

import java.io.*;
import java.util.*;

import weka.core.*;
import weka.classifiers.*;

/**
 * A SplitEvaluator that produces results for a classification scheme
 * on a nominal class attribute, including weighted misclassification costs.
 *
 * @author Len Trigg (len@reeltwo.com)
 * @version $Revision: 1.10.2.2 $
 */
public class CostSensitiveClassifierSplitEvaluator 
  extends ClassifierSplitEvaluator { 
  

  /** 
   * The directory used when loading cost files on demand, null indicates
   * current directory 
   */
  protected File m_OnDemandDirectory = new File(System.getProperty("user.dir"));

  /** The length of a result */
  private static final int RESULT_SIZE = 23;

  /**
   * Returns a string describing this split evaluator
   * @return a description of the split evaluator suitable for
   * displaying in the explorer/experimenter gui
   */
  public String globalInfo() {
    return weka.LocalString.get(" SplitEvaluator that produces results for a classification scheme ")
      +weka.LocalString.get("on a nominal class attribute, including weighted misclassification ")
      +weka.LocalString.get("costs.");
  }

  /**
   * Returns an enumeration describing the available options..
   *
   * @return an enumeration of all the available options.
   */
  public Enumeration listOptions() {

    Vector newVector = new Vector(1);
    Enumeration enu = super.listOptions();
    while (enu.hasMoreElements()) {
      newVector.addElement(enu.nextElement());
    }

    newVector.addElement(new Option(
              weka.LocalString.get("\tName of a directory to search for cost files when loading\n")
              +weka.LocalString.get("\tcosts on demand (default current directory)."),
              "D", 1, weka.LocalString.get("-D <directory>")));

    return newVector.elements();
  }

  /**
   * Parses a given list of options. Valid options (in addition to those of
   * ClassifierSplitEvaluator) are:<p>
   *
   * -D directory <br>
   * Name of a directory to search for cost files when loading costs on demand
   * (default current directory). <p>
   *
   * All option after -- will be passed to the classifier.
   *
   * @param options the list of options as an array of strings
   * @exception Exception if an option is not supported
   */
  public void setOptions(String[] options) throws Exception {
    
    String demandDir = Utils.getOption('D', options);
    if (demandDir.length() != 0) {
      setOnDemandDirectory(new File(demandDir));
    }

    super.setOptions(options);
  }

  /**
   * Gets the current settings of the Classifier.
   *
   * @return an array of strings suitable for passing to setOptions
   */
  public String [] getOptions() {

    String [] superOptions = super.getOptions();
    String [] options = new String [superOptions.length + 3];
    int current = 0;

    options[current++] = "-D";
    options[current++] = "" + getOnDemandDirectory();

    System.arraycopy(superOptions, 0, options, current, 
		     superOptions.length);
    current += superOptions.length;
    while (current < options.length) {
      options[current++] = "";
    }
    return options;
  }

  /**
   * Returns the tip text for this property
   * @return tip text for this property suitable for
   * displaying in the explorer/experimenter gui
   */
  public String onDemandDirectoryTipText() {
    return weka.LocalString.get("The directory to look in for cost files. This directory will be ")
      +weka.LocalString.get("searched for cost files when loading on demand.");
  }

  /**
   * Returns the directory that will be searched for cost files when
   * loading on demand.
   *
   * @return The cost file search directory.
   */
  public File getOnDemandDirectory() {

    return m_OnDemandDirectory;
  }

  /**
   * Sets the directory that will be searched for cost files when
   * loading on demand.
   *
   * @param newDir The cost file search directory.
   */
  public void setOnDemandDirectory(File newDir) {

    if (newDir.isDirectory()) {
      m_OnDemandDirectory = newDir;
    } else {
      m_OnDemandDirectory = new File(newDir.getParent());
    }
  }

  /**
   * Gets the data types of each of the result columns produced for a 
   * single run. The number of result fields must be constant
   * for a given SplitEvaluator.
   *
   * @return an array containing objects of the type of each result column. 
   * The objects should be Strings, or Doubles.
   */
  public Object [] getResultTypes() {
    int addm = (m_AdditionalMeasures != null) 
      ? m_AdditionalMeasures.length 
      : 0;
    Object [] resultTypes = new Object[RESULT_SIZE+addm];
    Double doub = new Double(0);
    int current = 0;
    resultTypes[current++] = doub;

    resultTypes[current++] = doub;
    resultTypes[current++] = doub;
    resultTypes[current++] = doub;
    resultTypes[current++] = doub;
    resultTypes[current++] = doub;
    resultTypes[current++] = doub;
    resultTypes[current++] = doub;
    resultTypes[current++] = doub;

    resultTypes[current++] = doub;
    resultTypes[current++] = doub;
    resultTypes[current++] = doub;
    resultTypes[current++] = doub;

    resultTypes[current++] = doub;
    resultTypes[current++] = doub;
    resultTypes[current++] = doub;
    resultTypes[current++] = doub;
    resultTypes[current++] = doub;
    resultTypes[current++] = doub;

    resultTypes[current++] = doub;
    resultTypes[current++] = doub;
    resultTypes[current++] = doub;

    resultTypes[current++] = "";

    // add any additional measures
    for (int i=0;i<addm;i++) {
      resultTypes[current++] = doub;
    }
    if (current != RESULT_SIZE+addm) {
      throw new Error(weka.LocalString.get("ResultTypes didn't fit RESULT_SIZE"));
    }
    return resultTypes;
  }

  /**
   * Gets the names of each of the result columns produced for a single run.
   * The number of result fields must be constant
   * for a given SplitEvaluator.
   *
   * @return an array containing the name of each result column
   */
  public String [] getResultNames() {
    int addm = (m_AdditionalMeasures != null) 
      ? m_AdditionalMeasures.length 
      : 0;
    String [] resultNames = new String[RESULT_SIZE+addm];
    int current = 0;
    resultNames[current++] = weka.LocalString.get("Number_of_instances");

    // Basic performance stats - right vs wrong
    resultNames[current++] = weka.LocalString.get("Number_correct");
    resultNames[current++] = weka.LocalString.get("Number_incorrect");
    resultNames[current++] = weka.LocalString.get("Number_unclassified");
    resultNames[current++] = weka.LocalString.get("Percent_correct");
    resultNames[current++] = weka.LocalString.get("Percent_incorrect");
    resultNames[current++] = weka.LocalString.get("Percent_unclassified");
    resultNames[current++] = weka.LocalString.get("Total_cost");
    resultNames[current++] = weka.LocalString.get("Average_cost");

    // Sensitive stats - certainty of predictions
    resultNames[current++] = weka.LocalString.get("Mean_absolute_error");
    resultNames[current++] = weka.LocalString.get("Root_mean_squared_error");
    resultNames[current++] = weka.LocalString.get("Relative_absolute_error");
    resultNames[current++] = weka.LocalString.get("Root_relative_squared_error");

    // SF stats
    resultNames[current++] = weka.LocalString.get("SF_prior_entropy");
    resultNames[current++] = weka.LocalString.get("SF_scheme_entropy");
    resultNames[current++] = weka.LocalString.get("SF_entropy_gain");
    resultNames[current++] = weka.LocalString.get("SF_mean_prior_entropy");
    resultNames[current++] = weka.LocalString.get("SF_mean_scheme_entropy");
    resultNames[current++] = weka.LocalString.get("SF_mean_entropy_gain");

    // K&B stats
    resultNames[current++] = weka.LocalString.get("KB_information");
    resultNames[current++] = weka.LocalString.get("KB_mean_information");
    resultNames[current++] = weka.LocalString.get("KB_relative_information");

    // Classifier defined extras
    resultNames[current++] = weka.LocalString.get("Summary");
    // add any additional measures
    for (int i=0;i<addm;i++) {
      resultNames[current++] = m_AdditionalMeasures[i];
    }
    if (current != RESULT_SIZE+addm) {
      throw new Error(weka.LocalString.get("ResultNames didn't fit RESULT_SIZE"));
    }
    return resultNames;
  }

  /**
   * Gets the results for the supplied train and test datasets. Now performs
   * a deep copy of the classifier before it is built and evaluated (just in case
   * the classifier is not initialized properly in buildClassifier()).
   *
   * @param train the training Instances.
   * @param test the testing Instances.
   * @return the results stored in an array. The objects stored in
   * the array may be Strings, Doubles, or null (for the missing value).
   * @exception Exception if a problem occurs while getting the results
   */
  public Object [] getResult(Instances train, Instances test) 
    throws Exception {

    if (train.classAttribute().type() != Attribute.NOMINAL) {
      throw new Exception(weka.LocalString.get("Class attribute is not nominal!"));
    }
    if (m_Template == null) {
      throw new Exception(weka.LocalString.get("No classifier has been specified"));
    }
    int addm = (m_AdditionalMeasures != null) 
      ? m_AdditionalMeasures.length 
      : 0;
    Object [] result = new Object[RESULT_SIZE+addm];

    String costName = train.relationName() + CostMatrix.FILE_EXTENSION;
    File costFile = new File(getOnDemandDirectory(), costName);
    if (!costFile.exists()) {
      throw new Exception(weka.LocalString.get("On-demand cost file doesn't exist: ") + costFile);
    }
    CostMatrix costMatrix = new CostMatrix(new BufferedReader(
                                           new FileReader(costFile)));

    Evaluation eval = new Evaluation(train, costMatrix);

    m_Classifier = Classifier.makeCopy(m_Template);
    m_Classifier.buildClassifier(train);
    eval.evaluateModel(m_Classifier, test);
    m_result = eval.toSummaryString();
    // The results stored are all per instance -- can be multiplied by the
    // number of instances to get absolute numbers
    int current = 0;
    result[current++] = new Double(eval.numInstances());

    result[current++] = new Double(eval.correct());
    result[current++] = new Double(eval.incorrect());
    result[current++] = new Double(eval.unclassified());
    result[current++] = new Double(eval.pctCorrect());
    result[current++] = new Double(eval.pctIncorrect());
    result[current++] = new Double(eval.pctUnclassified());
    result[current++] = new Double(eval.totalCost());
    result[current++] = new Double(eval.avgCost());

    result[current++] = new Double(eval.meanAbsoluteError());
    result[current++] = new Double(eval.rootMeanSquaredError());
    result[current++] = new Double(eval.relativeAbsoluteError());
    result[current++] = new Double(eval.rootRelativeSquaredError());

    result[current++] = new Double(eval.SFPriorEntropy());
    result[current++] = new Double(eval.SFSchemeEntropy());
    result[current++] = new Double(eval.SFEntropyGain());
    result[current++] = new Double(eval.SFMeanPriorEntropy());
    result[current++] = new Double(eval.SFMeanSchemeEntropy());
    result[current++] = new Double(eval.SFMeanEntropyGain());

    // K&B stats
    result[current++] = new Double(eval.KBInformation());
    result[current++] = new Double(eval.KBMeanInformation());
    result[current++] = new Double(eval.KBRelativeInformation());

    if (m_Classifier instanceof Summarizable) {
      result[current++] = ((Summarizable)m_Classifier).toSummaryString();
    } else {
      result[current++] = null;
    }
    
    for (int i=0;i<addm;i++) {
      if (m_doesProduce[i]) {
	try {
	  double dv = ((AdditionalMeasureProducer)m_Classifier).
	    getMeasure(m_AdditionalMeasures[i]);
	  if (!Instance.isMissingValue(dv)) {
	    Double value = new Double(dv);
	    result[current++] = value;
	  } else {
	    result[current++] = null;
	  }
	} catch (Exception ex) {
	  System.err.println(ex);
	}
      } else {
	result[current++] = null;
      }
    }

    if (current != RESULT_SIZE+addm) {
      throw new Error(weka.LocalString.get("Results didn't fit RESULT_SIZE"));
    }
    return result;
  }

  /**
   * Returns a text description of the split evaluator.
   *
   * @return a text description of the split evaluator.
   */
  public String toString() {

    String result = weka.LocalString.get("CostSensitiveClassifierSplitEvaluator: ");
    if (m_Template == null) {
      return result + weka.LocalString.get("<null> classifier");
    }
    return result + m_Template.getClass().getName() + " " 
      + m_ClassifierOptions + weka.LocalString.get("(version ") + m_ClassifierVersion + ")";
  }
} // CostSensitiveClassifierSplitEvaluator
